from dask.array.image import imread as daskread


def abspath(path):
    import os

    base_path = os.path.abspath(os.path.dirname(__file__))
    return os.path.join(base_path, path)


def get_test_data():

    image = daskread(abspath("data/carcinoma_xyzt.tif"))[0]
    return image


def get_stardist_modelpath():

    return abspath("models/Carcinoma_cells/")


def get_maskunet_modelpath():

    return abspath("models/Roi_Nuclei_Xenopus/")


def get_denoising_modelpath():

    return abspath("models/denoise_carcinoma/")



from .SmartSeeds3D import SmartSeeds3D
from .SmartSeeds2D import SmartSeeds2D
from .seedpool import SeedPool
from .unetstarmask import UnetStarMask
from .nmslabel import NMSLabel
from .OptimizeThreshold import OptimizeThreshold
from .pretrained import (
    register_model,
    register_aliases,
    clear_models_and_aliases,
)
from .UNET import UNET
from .MASKUNET import MASKUNET
from .StarDist2D import StarDist2D
from .StarDist3D import StarDist3D
from .Projection3D import Projection3D
from .CARE import CARE
from .SmartPatches import SmartPatches
from .SimplePatches import SimplePatches
from csbdeep.utils.tf import keras_import
from .CellPose import CellPose
from .PredictTiledLoader import PredictTiled
from .TrainTiledLoader import TrainTiled
from .Tiles_3D import VolumeSlicer
from .SmartNucleiPatches import SmartNucleiPatches
from .utils import (
    VollCellSeg,
    VollSeg,
    VollSeg2D,
    VollSeg3D,
    VollSeg_unet,
    VollSeg_nolabel_precondition,
    VollSeg_label_expansion,
    VollSeg_label_precondition,
    VollSeg_nolabel_expansion,
    merge_labels_across_volume,
    SimplePrediction,
    Skel,
    SmartSkel,
    STARPrediction3D,
    SuperSTARPrediction,
    SuperUNETPrediction,
    SuperWatershedwithMask,
    CellPose3DWater,
    CCLabels,
    CellPoseWater,
    CleanCellPose,
    CleanMask,
    VollOne,
)


get_file = keras_import("utils", "get_file")


__all__ = (
    "ResidualUNet3D",
    "SmartSeeds3D",
    "SmartSeeds2D",
    "SeedPool",
    "UnetStarMask",
    "NMSLabel",
    "OptimizeThreshold",
    "UNET",
    "MASKUNET",
    "StarDist2D",
    "StarDist3D",
    "Projection3D",
    "SimplePatches",
    "CARE",
    "SmartPatches",
    "CellPose",
    "UNet3D_cellpose",
    "PredictTiled",
    "CellPose3DModel",
    "TrainTiled",
    "CellPose3DTrain",
    "AbstractUNet",
    "UNet3D",
    "CellPose3DPredict",
    "VolumeSlicer",
    "VollCellSeg",
    "VollSeg",
    "VollSeg2D",
    "VollSeg3D",
    "VollSeg_unet",
    "VollSeg_nolabel_precondition",
    "VollSeg_label_expansion",
    "VollSeg_label_precondition",
    "VollSeg_nolabel_expansion",
    "merge_labels_across_volume",
    "SimplePrediction",
    "Skel",
    "SmartSkel",
    "STARPrediction3D",
    "SuperSTARPrediction",
    "SuperUNETPrediction",
    "SuperWatershedwithMask",
    "CellPose3DWater",
    "CCLabels",
    "CellPoseWater",
    "CleanCellPose",
    "CleanMask",
    "VollOne",
    "SmartNucleiPatches"
)

clear_models_and_aliases(StarDist2D, StarDist3D, UNET, CARE, MASKUNET)

register_model(
    CARE,
    "Denoise_3D_cells",
    "https://zenodo.org/record/6671170/files/GenericDenoising3D.zip",
    "a0eb25ffd794e2b3b31a4de5b72a392f",
)
register_model(
    CARE,
    "Denoise_carcinoma",
    "https://zenodo.org/record/5910645/files/denoise_carcinoma.zip",
    "fd33199738f0b17761272118cbffdf04",
)
register_model(
    UNET,
    "Embryo Cell Model (3D)",
    "https://zenodo.org/record/6337699/files/embryo_cell_model.zip",
    "c84fdec38a5b3cc6c1869c94ff23f3ba",
)
register_model(
    UNET,
    "Xenopus Tissue (2D)",
    "https://zenodo.org/record/6060378/files/Xenopus_tissue_model.zip",
    "2694d8b05fa828aceb055eef8cd5ca1f",
)
register_model(
    StarDist2D,
    "White_Blood_Cells",
    "https://zenodo.org/record/5815521/files/WBCSeg.zip",
    "7889f5902d8562766a4dee2726c90d49",
)
register_model(
    StarDist3D,
    "Carcinoma_cells",
    "https://zenodo.org/record/6354077/files/carcinoma_stardist.zip",
    "b92b9d5347862e52279629be575fe0b7",
)
register_model(
    UNET,
    "Microtubule Kymograph Segmentation",
    "https://zenodo.org/record/6355705/files/microtubule_kymograph_segmentation.zip",
    "a42fcd4ba732734d36eda3dbbb3d5673",
)
register_model(
    UNET,
    "Unet_White_Blood_Cells",
    "https://zenodo.org/record/5815588/files/UNETWBC.zip",
    "9645f004db478f661811d6da615ccc0b",
)
register_model(
    MASKUNET,
    "Xenopus_Cell_Tissue_Segmentation",
    "https://zenodo.org/record/6060378/files/Xenopus_tissue_model.zip",
    "2694d8b05fa828aceb055eef8cd5ca1f",
)
register_model(
    MASKUNET,
    "Unet_Arabidopsis_Mask",
    "https://zenodo.org/record/6670732/files/Unet_Arabidopsis_Mask.zip",
    "114df78e0153b39d80d0253a4dcc236f",
)
register_model(
    UNET,
    "Unet_Arabidopsis",
    "https://zenodo.org/record/6670747/files/Unet_Arabidopsis.zip",
    "ed7bdead6ebb11c3e13c22a156288f60",
)
register_model(
    UNET,
    "Unet_Cyto_White_Blood_Cells",
    "https://zenodo.org/record/5815603/files/UNETcytoWBC.zip",
    "dd3bf8b8e2a04536144954e882445a5e",
)
register_model(
    UNET,
    "Unet_Lung_Segmentation",
    "https://zenodo.org/record/6060177/files/Montgomery_county.zip",
    "be41937a00693e28961358440d242417",
)


register_aliases(UNET, "Embryo Cell Model (3D)", "Embryo Cell Model (3D)")
register_aliases(StarDist2D, "White_Blood_Cells", "White_Blood_Cells")
register_aliases(StarDist3D, "Carcinoma_cells", "Carcinoma_cells")
register_aliases(UNET, "Unet_White_Blood_Cells", "Unet_White_Blood_Cells")
register_aliases(UNET, "Unet_Cyto_White_Blood_Cells", "Unet_Cyto_White_Blood_Cells")
register_aliases(
    UNET,
    "Microtubule Kymograph Segmentation",
    "Microtubule Kymograph Segmentation",
)
register_aliases(UNET, "Xenopus Tissue (2D)", "Xenopus Tissue (2D)")
register_aliases(UNET, "Unet_Lung_Segmentation", "Unet_Lung_Segmentation")
register_aliases(UNET, "Unet_Arabidopsis", "Unet_Arabidopsis")
register_aliases(
    MASKUNET,
    "Xenopus_Cell_Tissue_Segmentation",
    "Xenopus_Cell_Tissue_Segmentation",
)
register_aliases(MASKUNET, "Unet_Arabidopsis_Mask", "Unet_Arabidopsis_Mask")
register_aliases(CARE, "Denoise_3D_cells", "Denoise_3D_cells")


del register_model, register_aliases, clear_models_and_aliases


def abspath(path):
    import os

    base_path = os.path.abspath(os.path.dirname(__file__))
    return os.path.join(base_path, path)


def test_image_ascadian_3d():
    from tifffile import imread

    url = "https://zenodo.org/record/5965906/files/Astec-Pm2_fuse_t001.tif"
    hash = "fdf1b78bc4ce4817000d1846db226118"

    img = imread(abspath(get_file(fname="Ascadian", origin=url, file_hash=hash)))
    return img


def test_image_arabidopsis_3d():
    from tifffile import imread

    url = "https://zenodo.org/record/6670569/files/04.tif"
    hash = "68204a6c871d6eeca9870728bfd1b8b7"

    img = imread(abspath(get_file(fname="Arabidopsis", origin=url, file_hash=hash)))
    return img


def test_image_carcinoma_3dt():
    from tifffile import imread

    url = "https://zenodo.org/record/6403439/files/carcinoma_xyzt.tif"
    hash = "713911848cf5263393e479d5cb3e5d59"
    img = imread(abspath(get_file(fname="Carcinoma", origin=url, file_hash=hash)))
    return img



#!/usr/bin/env python3
"""
Created on Mon Jan  3 19:36:44 2022

@author: varunkapoor
"""

from stardist.models import StarDist2D

from .pretrained import get_registered_models, get_model_details, get_model_instance
import sys


class StarDist2D(StarDist2D):
    def __init__(self, config, name=None, basedir="."):
        super().__init__(config=config, name=name, basedir=basedir)

    @classmethod
    def local_from_pretrained(cls, name_or_alias=None):
        try:
            print(cls)
            get_model_details(cls, name_or_alias, verbose=True)
            return get_model_instance(cls, name_or_alias)
        except ValueError:
            if name_or_alias is not None:
                print(
                    "Could not find model with name or alias '%s'" % (name_or_alias),
                    file=sys.stderr,
                )
                sys.stderr.flush()
            get_registered_models(cls, verbose=True)

    def predict_vollseg(
        self,
        img,
        axes=None,
        normalizer=None,
        prob_thresh=None,
        nms_thresh=None,
        n_tiles=None,
        show_tile_progress=True,
        verbose=False,
        return_labels=True,
        predict_kwargs=None,
        nms_kwargs=None,
        overlap_label=None,
        return_predict=False,
    ):
        """Predict instance segmentation from input image.
        Parameters
        ----------
        img : :class:`numpy.ndarray`
            Input image
        axes : str or None
            Axes of the input ``img``.
            ``None`` denotes that axes of img are the same as denoted in the config.
        normalizer : :class:`csbdeep.data.Normalizer` or None
            (Optional) normalization of input image before prediction.
            Note that the default (``None``) assumes ``img`` to be already normalized.
        sparse: bool
            If true, aggregate probabilities/distances sparsely during tiled
            prediction to save memory (recommended).
        prob_thresh : float or None
            Consider only object candidates from pixels with predicted object probability
            above this threshold (also see `optimize_thresholds`).
        nms_thresh : float or None
            Perform non-maximum suppression that considers two objects to be the same
            when their area/surface overlap exceeds this threshold (also see `optimize_thresholds`).
        n_tiles : iterable or None
            Out of memory (OOM) errors can occur if the input image is too large.
            To avoid this problem, the input image is broken up into (overlapping) tiles
            that are processed independently and re-assembled.
            This parameter denotes a tuple of the number of tiles for every image axis (see ``axes``).
            ``None`` denotes that no tiling should be used.
        show_tile_progress: bool
            Whether to show progress during tiled prediction.
        verbose: bool
            Whether to print some info messages.
        return_labels: bool
            Whether to create a label image, otherwise return None in its place.
        predict_kwargs: dict
            Keyword arguments for ``predict`` function of Keras model.
        nms_kwargs: dict
            Keyword arguments for non-maximum suppression.
        overlap_label: scalar or None
            if not None, label the regions where polygons overlap with that value
        return_predict: bool
            Also return the outputs of :func:`predict` (in a separate tuple)
            If True, implies sparse = False
        Returns
        -------
        (:class:`numpy.ndarray`, dict), (optional: return tuple of :func:`predict`)
            Returns a tuple of the label instances image and also
            a dictionary with the details (coordinates, etc.) of all remaining polygons/polyhedra.
        """
        if predict_kwargs is None:
            predict_kwargs = {}
        if nms_kwargs is None:
            nms_kwargs = {}

        nms_kwargs.setdefault("verbose", verbose)
        _axes = self._normalize_axes(img, axes)
        _axes_net = self.config.axes
        _permute_axes = self._make_permute_axes(_axes, _axes_net)
        _shape_inst = tuple(
            s for s, a in zip(_permute_axes(img).shape, _axes_net) if a != "C"
        )

        res = self.predict(
            img,
            axes=axes,
            normalizer=normalizer,
            n_tiles=n_tiles,
            show_tile_progress=show_tile_progress,
            **predict_kwargs
        )

        res = tuple(res) + (None,)
        prob, dist, points = res
        prob_class = None

        res_instances, polys = self._instances_from_prediction(
            _shape_inst,
            prob,
            dist,
            points=points,
            prob_class=prob_class,
            prob_thresh=prob_thresh,
            nms_thresh=nms_thresh,
            return_labels=return_labels,
            overlap_label=overlap_label,
            **nms_kwargs
        )

        return res_instances, prob, dist



import os
import numpy as np
from tifffile import imread, imwrite
from cellpose import models, metrics
import concurrent
from pathlib import Path


class CellPose:
    def __init__(
        self,
        base_dir,
        model_name,
        model_dir,
        raw_dir,
        real_mask_dir,
        test_raw_dir,
        test_real_mask_dir,
        n_epochs=400,
        diam_mean=30,
        learning_rate=0.0001,
        weight_decay=1.0e-4,
        channels=1,
        min_train_masks=1,
        gpu=True,
        real_train_3D=False,
        save_masks=True,
    ):

        self.base_dir = base_dir
        self.model_name = model_name
        self.model_dir = model_dir
        self.raw_dir = os.path.join(base_dir, raw_dir)
        self.real_mask_dir = os.path.join(base_dir, real_mask_dir)
        self.test_raw_dir = os.path.join(base_dir, test_raw_dir)
        self.test_real_mask_dir = os.path.join(base_dir, test_real_mask_dir)
        self.min_train_masks = min_train_masks
        self.save_raw_dir = os.path.join(
            self.base_dir, (raw_dir).replace("/", "") + "_sliced"
        )

        self.save_real_mask_dir = os.path.join(
            self.base_dir,
            (real_mask_dir).replace("/", "") + "_sliced",
        )
        self.save_test_raw_dir = os.path.join(
            self.base_dir,
            (test_raw_dir).replace("/", "") + "_sliced",
        )
        self.save_test_real_mask_dir = os.path.join(
            self.base_dir,
            (test_real_mask_dir).replace("/", "") + "_sliced",
        )

        self.n_epochs = n_epochs
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self.real_train_3D = real_train_3D
        self.channels = channels
        self.diam_mean = diam_mean
        self.save_masks = save_masks
        if model_dir and model_name is not None:
            self.pretrained_cellpose_model_path = os.path.join(
                model_dir, model_name
            )
        else:
            self.pretrained_cellpose_model_path = False
        self.gpu = gpu
        self.acceptable_formats = [".tif", ".TIFF", ".TIF", ".png"]
        if self.save_masks:
            Path(self.save_raw_dir).mkdir(exist_ok=True)
            Path(self.save_real_mask_dir).mkdir(exist_ok=True)
            Path(self.save_test_raw_dir).mkdir(exist_ok=True)
            Path(self.save_test_real_mask_dir).mkdir(exist_ok=True)

    def create_train(self):
        files_labels = os.listdir(self.real_mask_dir)
        (
            self.train_images,
            self.train_labels,
            self.train_names,
        ) = self._load_data(files_labels)

        files_test_labels = os.listdir(self.test_real_mask_dir)
        self.test_images, self.test_labels, self.test_names = self._load_data(
            files_test_labels
        )
        if self.save_masks:

            for i in range(len(self.train_images)):
                imwrite(
                    os.path.join(
                        self.save_raw_dir,
                        self.train_names[i] + "_" + str(i) + ".tif",
                    ),
                    self.train_images[i].astype(np.float32),
                )
                imwrite(
                    os.path.join(
                        self.save_real_mask_dir,
                        self.train_names[i] + "_" + str(i) + ".tif",
                    ),
                    self.train_labels[i].astype(np.uint16),
                )
            for i in range(len(self.test_images)):
                imwrite(
                    os.path.join(
                        self.save_test_raw_dir,
                        self.test_names[i] + "_" + str(i) + ".tif",
                    ),
                    self.test_images[i].astype(np.float32),
                )
                imwrite(
                    os.path.join(
                        self.save_test_real_mask_dir,
                        self.test_names[i] + "_" + str(i) + ".tif",
                    ),
                    self.test_labels[i].astype(np.uint16),
                )

    def get_data(self):
        files_labels = os.listdir(self.save_real_mask_dir)
        (
            self.train_images,
            self.train_labels,
            self.train_names,
        ) = self._load_saved_data(files_labels)

        files_test_labels = os.listdir(self.save_test_real_mask_dir)
        (
            self.test_images,
            self.test_labels,
            self.test_names,
        ) = self._load_saved_test_data(files_test_labels)

    def train(self):

        self.cellpose_model = models.CellposeModel(
            gpu=self.gpu,
            pretrained_model=self.pretrained_cellpose_model_path,
            diam_mean=self.diam_mean,
        )

        self.new_cellpose_model_path = self.cellpose_model.train(
            self.train_images,
            self.train_labels,
            test_data=self.test_images,
            test_labels=self.test_labels,
            save_path=self.model_dir,
            n_epochs=self.n_epochs,
            learning_rate=self.learning_rate,
            channels=self.channels,
            weight_decay=self.weight_decay,
            model_name=self.model_name,
            min_train_masks=self.min_train_masks,
        )
        self.diam_labels = self.cellpose_model.diam_labels.copy()

    def evaluate(self):

        self.masks = self.cellpose_model.eval(
            self.test_images, diameter=self.diam_labels
        )[0]
        ap = metrics.average_precision(self.test_labels, self.masks)[0]
        print("")
        print(
            f">>> average precision at iou threshold 0.5 = {ap[:,0].mean():.3f}"
        )

    def _load_saved_data(self, files_labels):

        images = []
        labels = []
        names = []
        for fname in files_labels:
            if any(fname.endswith(f) for f in self.acceptable_formats):
                name = os.path.splitext(fname)[0]
                labelimage = imread(
                    os.path.join(self.save_real_mask_dir, fname)
                ).astype(np.uint16)
                image = imread(os.path.join(self.save_raw_dir, fname))

                labels.append(labelimage)
                images.append(image)
                names.append(name)

        return images, labels, names

    def _load_saved_test_data(self, files_labels):

        images = []
        labels = []
        names = []
        for fname in files_labels:
            if any(fname.endswith(f) for f in self.acceptable_formats):
                name = os.path.splitext(fname)[0]
                labelimage = imread(
                    os.path.join(self.save_test_real_mask_dir, fname)
                ).astype(np.uint16)
                image = imread(os.path.join(self.save_test_raw_dir, fname))

                labels.append(labelimage)
                images.append(image)
                names.append(name)

        return images, labels, names

    def _load_data(self, files_labels):

        images = []
        labels = []
        names = []
        for fname in files_labels:
            if any(fname.endswith(f) for f in self.acceptable_formats):
                name = os.path.splitext(fname)[0]
                labelimage = imread(
                    os.path.join(self.real_mask_dir, fname)
                ).astype(np.uint16)
                image = imread(os.path.join(self.raw_dir, fname))
                if not self.real_train_3D:
                    with concurrent.futures.ThreadPoolExecutor(
                        max_workers=os.cpu_count() - 1
                    ) as executor:
                        future_labels = [
                            executor.submit(slicer, labelimage, i)
                            for i in range(labelimage.shape[0])
                        ]
                        future_raw = [
                            executor.submit(slicer, image, i)
                            for i in range(image.shape[0])
                        ]
                        current_labels = [
                            r.result()
                            for r in concurrent.futures.as_completed(
                                future_labels
                            )
                        ]
                        current_raw = [
                            r.result()
                            for r in concurrent.futures.as_completed(
                                future_raw
                            )
                        ]
                    for i in range(len(current_labels)):
                        if current_labels[i].max() > 0:
                            labels.append(current_labels[i])
                            images.append(current_raw[i])
                            current_name = name + str(i)
                            names.append(current_name)

                else:
                    labels.append(labelimage)
                    images.append(image)
                    names.append(name)

        return images, labels, names


def slicer(image, i):

    return image[i, :]



import os
import numpy as np
from pathlib import Path
from tifffile import imread, imwrite
from skimage.measure import regionprops
from skimage.morphology import remove_small_objects
from scipy.ndimage import binary_erosion
from tqdm import tqdm
from datetime import datetime
from uuid import uuid4
from skimage.segmentation import find_boundaries
from skimage.morphology import dilation
from skimage.filters import gaussian


class SmartNucleiPatches:
    def __init__(
        self,
        base_signal_dir,
        raw_signal_dir,
        signal_channel_results_directory,
        signal_raw_save_dir,
        signal_real_mask_patch_dir,
        signal_binary_mask_patch_dir,
        patch_size,
        erosion_iterations=0,
        pattern=".tif",
        lower_ratio_fore_to_back=0.5,
        upper_ratio_fore_to_back=0.9,
        max_foreground_patches_per_image=np.inf,
        max_background_patches_per_image=np.inf,
        create_background_only=False,
        create_foreground_only=False,
    ):

        self.max_foreground_patches_per_image = (
            max_foreground_patches_per_image
        )
        self.max_background_patches_per_image = (
            max_background_patches_per_image
        )
        self.base_signal_dir = base_signal_dir
        self.create_background_only = create_background_only
        self.create_foreground_only = create_foreground_only
        
        self.raw_signal_dir = os.path.join(
            self.base_signal_dir, raw_signal_dir
        )
        
        self.signal_channel_results_directory = os.path.join(
            self.base_signal_dir, signal_channel_results_directory
        )
        self.signal_raw_save_dir = os.path.join(
            base_signal_dir, signal_raw_save_dir
        )
        
        self.signal_binary_mask_patch_dir = os.path.join(
            self.base_signal_dir, signal_binary_mask_patch_dir
        )


        self.signal_real_mask_patch_dir = os.path.join(
            self.base_signal_dir, signal_real_mask_patch_dir
        )
        
        self.patch_size = patch_size
        self.erosion_iterations = erosion_iterations
        self.pattern = pattern
        self.lower_ratio_fore_to_back = lower_ratio_fore_to_back
        self.upper_ratio_fore_to_back = upper_ratio_fore_to_back
        self.acceptable_formats = [".tif", ".TIFF", ".TIF", ".png"]

        if self.create_background_only:
            self._create_background_patches()
        elif self.create_foreground_only:
            self._create_smart_patches()
        else:
            self._create_smart_patches()
            self._create_background_patches()

    def _create_background_patches(self):

        Path(self.signal_raw_save_dir).mkdir(exist_ok=True)
        Path(self.signal_binary_mask_patch_dir).mkdir(exist_ok=True)
        Path(self.signal_real_mask_patch_dir).mkdir(exist_ok=True)
        files = os.listdir(self.raw_signal_dir)
        
        for fname in files:
            if any(fname.endswith(f) for f in self.acceptable_formats):

                    self.main_count = 0

                    raw_signal_image = imread(
                        os.path.join(self.raw_signal_dir, fname)
                    )
                    self.ndim = len(raw_signal_image.shape)
                    label_image_signal = imread(
                        os.path.join(
                            self.signal_channel_results_directory, fname
                        )
                    ).astype(np.uint16)

                    self._background_label_maker(
                        fname,
                        raw_signal_image,
                        label_image_signal,
                        self.signal_raw_save_dir,
                        self.signal_binary_mask_patch_dir,
                        self.signal_real_mask_patch_dir,
                    )

            

    def _create_smart_patches(self):

        Path(self.signal_raw_save_dir).mkdir(exist_ok=True)
        Path(self.signal_binary_mask_patch_dir).mkdir(exist_ok=True)
        Path(self.signal_real_mask_patch_dir).mkdir(exist_ok=True)
        files = os.listdir(self.raw_signal_dir)
       
        for fname in files:
            if any(fname.endswith(f) for f in self.acceptable_formats):

                    self.main_count = 0

                    raw_signal_image = imread(
                        os.path.join(self.raw_signal_dir, fname)
                    )
                    self.ndim = len(raw_signal_image.shape)
                    label_image_signal = imread(
                        os.path.join(
                            self.signal_channel_results_directory, fname
                        )
                    ).astype(np.uint16)
                    properties_signal = regionprops(label_image_signal)
                    for count, prop in tqdm(enumerate(properties_signal)):
                        self._label_maker(
                            fname,
                            raw_signal_image,
                            label_image_signal,
                            count,
                            prop,
                            self.signal_raw_save_dir,
                            self.signal_binary_mask_patch_dir,
                            self.signal_real_mask_patch_dir,
                        )
                

    def _background_label_maker(
        self,
        fname,
        rawimage,
        labelimage,
        raw_save_dir: str,
        binary_mask_patch_dir: str,
        real_mask_patch_dir: str,
    ):

        zero_indices = list(zip(*np.where(labelimage == 0)))
        self.main_count = 0
        for index in zero_indices:

            if self.main_count < self.max_background_patches_per_image:
                name = os.path.splitext(fname)[0]
                if self.ndim == 2:
                    x = index[1]
                    y = index[0]
                    crop_Xminus = x - int(self.patch_size[1] / 2)
                    crop_Xplus = x + int(self.patch_size[1] / 2)
                    crop_Yminus = y - int(self.patch_size[0] / 2)
                    crop_Yplus = y + int(self.patch_size[0] / 2)
                    if (
                        crop_Xminus > 0
                        and crop_Xplus < rawimage.shape[1]
                        and crop_Yminus > 0
                        and crop_Yplus < rawimage.shape[0]
                    ):

                        properties = regionprops(labelimage)
                        for prop in properties:
                            centroid = prop.centroid
                            xc = centroid[1]
                            yc = centroid[0]
                            crop_Xminusc = xc - int(self.patch_size[1] / 2)
                            crop_Xplusc = xc + int(self.patch_size[1] / 2)
                            crop_Yminusc = yc - int(self.patch_size[0] / 2)
                            crop_Yplusc = yc + int(self.patch_size[0] / 2)
                            regionc = (
                                slice(int(crop_Yminusc), int(crop_Yplusc)),
                                slice(int(crop_Xminusc), int(crop_Xplusc)),
                            )

                            if (
                                crop_Xminusc > 0
                                and crop_Xplusc < rawimage.shape[1]
                                and crop_Yminusc > 0
                                and crop_Yplusc < rawimage.shape[0]
                            ):
                                raw_patch = rawimage[
                                    crop_Yminus:crop_Yplus,
                                    crop_Xminus:crop_Xplus,
                                ]
                                mask_patch_zero = labelimage[
                                    crop_Yminus:crop_Yplus,
                                    crop_Xminus:crop_Xplus,
                                ]

                                raw_patchc = rawimage[regionc]
                                mask_patchc = labelimage[regionc]
                                raw_patch = np.add(raw_patch, raw_patchc)
                                mask_patch = np.add(
                                    mask_patch_zero, mask_patchc
                                )

                                if (
                                    np.sum(raw_patch) > 0
                                    and np.sum(mask_patch_zero) == 0
                                ):
                                    self.main_count += 1
                                    eventid = datetime.now().strftime(
                                        "%Y%m-%d%H-%M%S-"
                                    ) + str(uuid4())

                                    imwrite(
                                        os.path.join(
                                            raw_save_dir,
                                            name
                                            + "back"
                                            + eventid
                                            + str(self.main_count)
                                            + ".tif",
                                        ),
                                        raw_patch.astype("float32"),
                                    )
                                    if self.erosion_iterations > 0:
                                        binary_mask_patch = erode_labels(
                                            mask_patch.astype("uint16"),
                                            self.erosion_iterations,
                                        )
                                    else:
                                        binary_mask_patch = mask_patch
                                    binary_mask_patch = binary_mask_patch > 0
                                    imwrite(
                                        os.path.join(
                                            binary_mask_patch_dir,
                                            name
                                            + "back"
                                            + eventid
                                            + str(self.main_count)
                                            + ".tif",
                                        ),
                                        binary_mask_patch.astype("uint16"),
                                    )

                                    imwrite(
                                        os.path.join(
                                            real_mask_patch_dir,
                                            name
                                            + "back"
                                            + eventid
                                            + str(self.main_count)
                                            + ".tif",
                                        ),
                                        mask_patch.astype("uint16"),
                                    )
                if self.ndim == 3:
                    x = index[2]
                    y = index[1]
                    z = index[0]
                    crop_Xminus = x - int(self.patch_size[2] / 2)
                    crop_Xplus = x + int(self.patch_size[2] / 2)
                    crop_Yminus = y - int(self.patch_size[1] / 2)
                    crop_Yplus = y + int(self.patch_size[1] / 2)
                    crop_Zminus = z - int(self.patch_size[0] / 2)
                    crop_Zplus = z + int(self.patch_size[0] / 2)
                    if (
                        crop_Xminus > 0
                        and crop_Xplus < rawimage.shape[2]
                        and crop_Yminus > 0
                        and crop_Yplus < rawimage.shape[1]
                        and crop_Zminus > 0
                        and crop_Zplus < rawimage.shape[0]
                    ):
                        properties = regionprops(labelimage)
                        for prop in properties:
                            centroid = prop.centroid
                            xc = centroid[2]
                            yc = centroid[1]
                            zc = centroid[0]

                            crop_Xminusc = xc - int(self.patch_size[2] / 2)
                            crop_Xplusc = xc + int(self.patch_size[2] / 2)
                            crop_Yminusc = yc - int(self.patch_size[1] / 2)
                            crop_Yplusc = yc + int(self.patch_size[1] / 2)
                            crop_Zminusc = zc - int(self.patch_size[0] / 2)
                            crop_Zplusc = zc + int(self.patch_size[0] / 2)
                            regionc = (
                                slice(int(crop_Zminusc), int(crop_Zplusc)),
                                slice(int(crop_Yminusc), int(crop_Yplusc)),
                                slice(int(crop_Xminusc), int(crop_Xplusc)),
                            )

                            if (
                                crop_Xminusc > 0
                                and crop_Xplusc < rawimage.shape[2]
                                and crop_Yminusc > 0
                                and crop_Yplusc < rawimage.shape[1]
                                and crop_Zminusc > 0
                                and crop_Zplusc < rawimage.shape[0]
                            ):
                                raw_patch = rawimage[
                                    crop_Zminus:crop_Zplus,
                                    crop_Yminus:crop_Yplus,
                                    crop_Xminus:crop_Xplus,
                                ]
                                mask_patch_zero = labelimage[
                                    crop_Zminus:crop_Zplus,
                                    crop_Yminus:crop_Yplus,
                                    crop_Xminus:crop_Xplus,
                                ]

                                raw_patchc = rawimage[regionc]
                                mask_patchc = labelimage[regionc]
                                raw_patch = np.add(raw_patch, raw_patchc)
                                mask_patch = np.add(
                                    mask_patch_zero, mask_patchc
                                )

                                if (
                                    np.sum(raw_patch) > 0
                                    and np.sum(mask_patch_zero) == 0
                                ):
                                    self.main_count += 1
                                    eventid = datetime.now().strftime(
                                        "%Y%m-%d%H-%M%S-"
                                    ) + str(uuid4())

                                    imwrite(
                                        os.path.join(
                                            raw_save_dir,
                                            name
                                            + "back"
                                            + eventid
                                            + str(self.main_count)
                                            + ".tif",
                                        ),
                                        raw_patch.astype("float32"),
                                    )
                                    if self.erosion_iterations > 0:
                                        binary_mask_patch = erode_labels(
                                            mask_patch.astype("uint16"),
                                            self.erosion_iterations,
                                        )
                                    else:
                                        binary_mask_patch = mask_patch
                                    binary_mask_patch = binary_mask_patch > 0
                                    imwrite(
                                        os.path.join(
                                            binary_mask_patch_dir,
                                            name
                                            + "back"
                                            + eventid
                                            + str(self.main_count)
                                            + ".tif",
                                        ),
                                        binary_mask_patch.astype("uint16"),
                                    )

                                    imwrite(
                                        os.path.join(
                                            real_mask_patch_dir,
                                            name
                                            + "back"
                                            + eventid
                                            + str(self.main_count)
                                            + ".tif",
                                        ),
                                        mask_patch.astype("uint16"),
                                    )

    def _label_maker(
        self,
        fname: str,
        rawimage: np.ndarray,
        labelimage: np.ndarray,
        count: int,
        prop: regionprops,
        raw_save_dir: str,
        binary_mask_patch_dir: str,
        real_mask_patch_dir: str,
    ):

        name = os.path.splitext(fname)[0]

        if self.ndim == 2:

            self.valid = False
            centroid = prop.centroid
            x = centroid[1]
            y = centroid[0]

            crop_Xminus = x - int(self.patch_size[1] / 2)
            crop_Xplus = x + int(self.patch_size[1] / 2)
            crop_Yminus = y - int(self.patch_size[0] / 2)
            crop_Yplus = y + int(self.patch_size[0] / 2)
            crop_minus = [crop_Yminus, crop_Xminus]
            region = (
                slice(int(crop_Yminus), int(crop_Yplus)),
                slice(int(crop_Xminus), int(crop_Xplus)),
            )
        if self.ndim == 3:

            self.valid = False
            centroid = prop.centroid
            z = centroid[0]
            x = centroid[2]
            y = centroid[1]

            crop_Xminus = x - int(self.patch_size[2] / 2)
            crop_Xplus = x + int(self.patch_size[2] / 2)
            crop_Yminus = y - int(self.patch_size[1] / 2)
            crop_Yplus = y + int(self.patch_size[1] / 2)
            crop_Zminus = z - int(self.patch_size[0] / 2)
            crop_Zplus = z + int(self.patch_size[0] / 2)
            crop_minus = [crop_Zminus, crop_Yminus, crop_Xminus]
            region = (
                slice(int(crop_Zminus), int(crop_Zplus)),
                slice(int(crop_Yminus), int(crop_Yplus)),
                slice(int(crop_Xminus), int(crop_Xplus)),
            )
        if all(crop for crop in crop_minus) > 0:
            self.crop_labelimage = labelimage[region]
            self.crop_image = rawimage[region]
            self.crop_labelimage = remove_small_objects(
                self.crop_labelimage.astype("uint16"), min_size=10
            )
            if (
                self.crop_labelimage.shape[0] == self.patch_size[0]
                and self.crop_labelimage.shape[1] == self.patch_size[1]
                and self.ndim == 2
            ):
                self._crop_maker(
                    name,
                    count,
                    raw_save_dir,
                    binary_mask_patch_dir,
                    real_mask_patch_dir,
                )
            if (
                self.crop_labelimage.shape[0] == self.patch_size[0]
                and self.crop_labelimage.shape[1] == self.patch_size[1]
                and self.crop_labelimage.shape[2] == self.patch_size[2]
                and self.ndim == 3
            ):
                self._crop_maker(
                    name,
                    count,
                    raw_save_dir,
                    binary_mask_patch_dir,
                    real_mask_patch_dir,
                )

    def _crop_maker(
        self,
        name,
        count,
        raw_save_dir,
        binary_mask_patch_dir,
        real_mask_patch_dir,
    ):

        self._region_selector()
        if (
            self.valid
            and self.main_count < self.max_foreground_patches_per_image
        ):
            self.main_count += 1
            if self.erosion_iterations > 0:
                self.eroded_crop_labelimage = erode_labels(
                    self.crop_labelimage.astype("uint16"),
                    self.erosion_iterations,
                )
            else:
                self.eroded_crop_labelimage = self.crop_labelimage
            eroded_binary_image = self.eroded_crop_labelimage > 0
            eventid = datetime.now().strftime("%Y%m-%d%H-%M%S-") + str(uuid4())
            imwrite(
                os.path.join(
                    binary_mask_patch_dir,
                    name + eventid + str(count) + self.pattern,
                ),
                eroded_binary_image.astype("uint16"),
            )
            imwrite(
                os.path.join(
                    raw_save_dir, name + eventid + str(count) + self.pattern
                ),
                self.crop_image.astype("float32"),
            )
            imwrite(
                os.path.join(
                    real_mask_patch_dir,
                    name + eventid + str(count) + self.pattern,
                ),
                self.eroded_crop_labelimage.astype("uint16"),
            )

    def _region_selector(self):

        non_zero_indices = list(zip(*np.where(self.crop_labelimage > 0)))

        total_indices = list(zip(*np.where(self.crop_labelimage >= 0)))
        if len(total_indices) > 0:
            norm_foreground = len(non_zero_indices) / len(total_indices)
            index_ratio = float(norm_foreground)
            if (
                index_ratio >= self.lower_ratio_fore_to_back
                and index_ratio <= self.upper_ratio_fore_to_back
            ):

                self.valid = True


def erode_label_holes(lbl_img, iterations):
    lbl_img_filled = np.zeros_like(lbl_img)
    for lb in range(np.min(lbl_img), np.max(lbl_img) + 1):
        mask = lbl_img == lb
        mask_filled = binary_erosion(mask, iterations=iterations)
        lbl_img_filled[mask_filled] = lb
    return lbl_img_filled


def erode_labels(segmentation, erosion_iterations=2):
    # create empty list where the eroded masks can be saved to
    regions = regionprops(segmentation)
    erode = np.zeros(segmentation.shape)

    def erode_mask(segmentation_labels, label_id, erosion_iterations):

        only_current_label_id = np.where(segmentation_labels == label_id, 1, 0)
        eroded = binary_erosion(
            only_current_label_id, iterations=erosion_iterations
        )
        relabeled_eroded = np.where(eroded == 1, label_id, 0)
        return relabeled_eroded

    for i in range(len(regions)):
        label_id = regions[i].label
        erode = erode + erode_mask(segmentation, label_id, erosion_iterations)

    # convert list of numpy arrays to stacked numpy array
    return erode



import math
import typing
from typing import List, Iterable, Tuple

import cv2
import numpy as np
import torch


def compute_pyramid_patch_weight_loss(width: int, height: int) -> np.ndarray:
    """Compute a weight matrix that assigns bigger weight on pixels in center and
    less weight to pixels on image boundary.
    This weight matrix then used for merging individual tile predictions and helps dealing
    with prediction artifacts on tile boundaries.
    :param width: Tile width
    :param height: Tile height
    :return: Since-channel image [Width x Height]
    """
    xc = width * 0.5
    yc = height * 0.5
    xl = 0
    xr = width
    yb = 0
    yt = height
    Dc = np.zeros((width, height))
    De = np.zeros((width, height))

    Dcx = np.square(np.arange(width) - xc + 0.5)
    Dcy = np.square(np.arange(height) - yc + 0.5)
    Dc = np.sqrt(Dcx[np.newaxis].transpose() + Dcy)

    De_l = np.square(np.arange(width) - xl + 0.5) + np.square(0.5)
    De_r = np.square(np.arange(width) - xr + 0.5) + np.square(0.5)
    De_b = np.square(0.5) + np.square(np.arange(height) - yb + 0.5)
    De_t = np.square(0.5) + np.square(np.arange(height) - yt + 0.5)

    De_x = np.sqrt(np.minimum(De_l, De_r))
    De_y = np.sqrt(np.minimum(De_b, De_t))
    De = np.minimum(De_x[np.newaxis].transpose(), De_y)

    alpha = (width * height) / np.sum(np.divide(De, np.add(Dc, De)))
    W = alpha * np.divide(De, np.add(Dc, De))
    return W, Dc, De


class ImageSlicer:
    tile_size: Tuple[int, int]
    tile_step: Tuple[int, int]
    overlap: Tuple[int, int]

    """
    Helper class to slice image into tiles and merge them back
    """

    def __init__(
        self,
        image_shape: Tuple[int, int],
        tile_size,
        tile_step=0,
        image_margin=0,
        weight="mean",
    ):
        """
        :param image_shape: Shape of the source image (H, W)
        :param tile_size: Tile size (Scalar or tuple (H, W)
        :param tile_step: Step in pixels between tiles (Scalar or tuple (H, W))
        :param image_margin:
        :param weight: Fusion algorithm. 'mean' - avergaing
        """
        self.image_height = image_shape[0]
        self.image_width = image_shape[1]

        if isinstance(tile_size, (np.ndarray, typing.Sequence)):
            if len(tile_size) != 2:
                raise ValueError(
                    f"Tile size must have exactly 2 elements. Got: tile_size={tile_size}"
                )
            self.tile_size = int(tile_size[0]), int(tile_size[1])
        else:
            self.tile_size = int(tile_size), int(tile_size)

        if isinstance(tile_step, (np.ndarray, typing.Sequence)):
            if len(tile_step) != 2:
                raise ValueError(
                    f"Tile size must have exactly 2 elements. Got: tile_step={tile_size}"
                )
            self.tile_step = int(tile_step[0]), int(tile_step[1])
        else:
            self.tile_step = int(tile_step), int(tile_step)

        weights = {"mean": self._mean, "pyramid": self._pyramid}

        self.weight = (
            weight
            if isinstance(weight, np.ndarray)
            else weights[weight](self.tile_size)
        )

        if self.tile_step[0] < 1 or self.tile_step[0] > self.tile_size[0]:
            raise ValueError()
        if self.tile_step[1] < 1 or self.tile_step[1] > self.tile_size[1]:
            raise ValueError()

        overlap = (
            self.tile_size[0] - self.tile_step[0],
            self.tile_size[1] - self.tile_step[1],
        )

        self.margin_left = 0
        self.margin_right = 0
        self.margin_top = 0
        self.margin_bottom = 0

        if image_margin == 0:
            # In case margin is not set, we compute it manually

            nw = max(
                1,
                math.ceil((self.image_width - overlap[1]) / self.tile_step[1]),
            )
            nh = max(
                1,
                math.ceil(
                    (self.image_height - overlap[0]) / self.tile_step[0]
                ),
            )

            extra_w = self.tile_step[1] * nw - (self.image_width - overlap[1])
            extra_h = self.tile_step[0] * nh - (self.image_height - overlap[0])

            self.margin_left = extra_w // 2
            self.margin_right = extra_w - self.margin_left
            self.margin_top = extra_h // 2
            self.margin_bottom = extra_h - self.margin_top

        else:
            if isinstance(image_margin, typing.Sequence):
                (
                    margin_left,
                    margin_right,
                    margin_top,
                    margin_bottom,
                ) = image_margin
            else:
                margin_left = (
                    margin_right
                ) = margin_top = margin_bottom = image_margin

            self.margin_left = margin_left
            self.margin_right = margin_right
            self.margin_top = margin_top
            self.margin_bottom = margin_bottom

        crops = []
        bbox_crops = []

        for y in range(
            0,
            self.image_height
            + self.margin_top
            + self.margin_bottom
            - self.tile_size[0]
            + 1,
            self.tile_step[0],
        ):
            for x in range(
                0,
                self.image_width
                + self.margin_left
                + self.margin_right
                - self.tile_size[1]
                + 1,
                self.tile_step[1],
            ):
                crops.append((x, y, self.tile_size[1], self.tile_size[0]))
                bbox_crops.append(
                    (
                        x - self.margin_left,
                        y - self.margin_top,
                        self.tile_size[1],
                        self.tile_size[0],
                    )
                )

        self.crops = np.array(crops)
        self.bbox_crops = np.array(bbox_crops)

    def iter_split(
        self, image: np.ndarray, border_type=cv2.BORDER_CONSTANT, value=0
    ) -> Iterable[Tuple[np.ndarray, Tuple[int, int, int, int]]]:
        if (image.shape[0] != self.image_height) or (
            image.shape[1] != self.image_width
        ):
            raise ValueError()

        orig_shape_len = len(image.shape)

        for coords, crop_coords in zip(self.crops, self.bbox_crops):
            x, y, tile_width, tile_height = crop_coords
            x1 = max(x, 0)
            y1 = max(y, 0)
            x2 = min(image.shape[1], x + tile_width)
            y2 = min(image.shape[0], y + tile_height)

            tile = image[y1:y2, x1:x2]  # .copy()
            if (
                x < 0
                or y < 0
                or (x + tile_width) > image.shape[1]
                or (y + tile_height) > image.shape[0]
            ):
                tile = cv2.copyMakeBorder(
                    tile,
                    top=max(0, -y),
                    bottom=max(0, y + tile_height - image.shape[0]),
                    left=max(0, -x),
                    right=max(0, x + tile_width - image.shape[1]),
                    borderType=border_type,
                    value=value,
                )

                # This check recovers possible lack of last dummy dimension for single-channel images
                if len(tile.shape) != orig_shape_len:
                    tile = np.expand_dims(tile, axis=-1)

            yield tile, coords

    def split(self, image, border_type=cv2.BORDER_CONSTANT, value=0):
        assert image.shape[0] == self.image_height
        assert image.shape[1] == self.image_width

        orig_shape_len = len(image.shape)
        image = cv2.copyMakeBorder(
            image,
            self.margin_top,
            self.margin_bottom,
            self.margin_left,
            self.margin_right,
            borderType=border_type,
            value=value,
        )

        # This check recovers possible lack of last dummy dimension for single-channel images
        if len(image.shape) != orig_shape_len:
            image = np.expand_dims(image, axis=-1)

        tiles = []
        for x, y, tile_width, tile_height in self.crops:
            tile = image[y : y + tile_height, x : x + tile_width]  # .copy()
            assert tile.shape[0] == self.tile_size[0]
            assert tile.shape[1] == self.tile_size[1]

            tiles.append(tile)

        return tiles

    def cut_patch(
        self,
        image: np.ndarray,
        slice_index,
        border_type=cv2.BORDER_CONSTANT,
        value=0,
    ):
        assert image.shape[0] == self.image_height
        assert image.shape[1] == self.image_width

        orig_shape_len = len(image.shape)
        x, y, tile_width, tile_height = self.bbox_crops[slice_index]

        x1 = max(x, 0)
        y1 = max(y, 0)
        x2 = min(image.shape[1], x + tile_width)
        y2 = min(image.shape[0], y + tile_height)

        tile = image[y1:y2, x1:x2]
        if (
            x < 0
            or y < 0
            or (x + tile_width) > image.shape[1]
            or (y + tile_height) > image.shape[0]
        ):
            tile = cv2.copyMakeBorder(
                tile,
                top=max(0, -y),
                bottom=max(0, y + tile_height - image.shape[0]),
                left=max(0, -x),
                right=max(0, x + tile_width - image.shape[1]),
                borderType=border_type,
                value=value,
            )

            # This check recovers possible lack of last dummy dimension for single-channel images
            if len(tile.shape) != orig_shape_len:
                tile = np.expand_dims(tile, axis=-1)

        return tile

    @property
    def target_shape(self):
        target_shape = (
            self.image_height + self.margin_bottom + self.margin_top,
            self.image_width + self.margin_right + self.margin_left,
        )
        return target_shape

    def merge(self, tiles: List[np.ndarray], dtype=np.float32):
        if len(tiles) != len(self.crops):
            raise ValueError

        channels = 1 if len(tiles[0].shape) == 2 else tiles[0].shape[2]
        target_shape = (
            self.image_height + self.margin_bottom + self.margin_top,
            self.image_width + self.margin_right + self.margin_left,
            channels,
        )

        image = np.zeros(target_shape, dtype=np.float64)
        norm_mask = np.zeros(target_shape, dtype=np.float64)

        w = np.dstack([self.weight] * channels)

        for tile, (x, y, tile_width, tile_height) in zip(tiles, self.crops):
            # print(x, y, tile_width, tile_height, image.shape)
            image[y : y + tile_height, x : x + tile_width] += tile * w
            norm_mask[y : y + tile_height, x : x + tile_width] += w

        # print(norm_mask.min(), norm_mask.max())
        norm_mask = np.clip(
            norm_mask, a_min=np.finfo(norm_mask.dtype).eps, a_max=None
        )
        normalized = np.divide(image, norm_mask).astype(dtype)
        crop = self.crop_to_orignal_size(normalized)
        return crop

    def crop_to_orignal_size(self, image):
        assert image.shape[0] == self.target_shape[0]
        assert image.shape[1] == self.target_shape[1]
        crop = image[
            self.margin_top : self.image_height + self.margin_top,
            self.margin_left : self.image_width + self.margin_left,
        ]
        assert crop.shape[0] == self.image_height
        assert crop.shape[1] == self.image_width
        return crop

    def _mean(self, tile_size):
        return np.ones((tile_size[0], tile_size[1]), dtype=np.float32)

    def _pyramid(self, tile_size):
        w, _, _ = compute_pyramid_patch_weight_loss(tile_size[0], tile_size[1])
        return w


class TileMerger:
    """
    Helper class to merge final image on GPU. This generally faster than moving individual tiles to CPU.
    """

    def __init__(
        self, image_shape, channels, weight, device="cpu", dtype=torch.float32
    ):
        """
        :param image_shape: Shape of the source image
        :param image_margin:
        :param weight: Weighting matrix
        """
        self.image_height = image_shape[0]
        self.image_width = image_shape[1]
        self.channels = channels

        self.weight = torch.from_numpy(np.expand_dims(weight, axis=0)).to(
            device=device, dtype=dtype
        )
        self.image = torch.zeros(
            (channels, self.image_height, self.image_width),
            device=device,
            dtype=dtype,
        )
        self.norm_mask = torch.zeros(
            (1, self.image_height, self.image_width),
            device=device,
            dtype=dtype,
        )

    def accumulate_single(self, tile: torch.Tensor, coords):
        """
        Accumulates single element
        :param sample: Predicted image of shape [C,H,W]
        :param crop_coords: Corresponding tile crops w.r.t to original image
        """
        tile = tile.to(device=self.image.device)
        x, y, tile_width, tile_height = coords
        self.image[:, y : y + tile_height, x : x + tile_width] += (
            tile * self.weight
        )
        self.norm_mask[
            :, y : y + tile_height, x : x + tile_width
        ] += self.weight

    def integrate_batch(self, batch: torch.Tensor, crop_coords):
        """
        Accumulates batch of tile predictions
        :param batch: Predicted tiles  of shape [B,C,H,W]
        :param crop_coords: Corresponding tile crops w.r.t to original image
        """
        if len(batch) != len(crop_coords):
            raise ValueError(
                "Number of images in batch does not correspond to number of coordinates"
            )

        if batch.device != self.image.device:
            batch = batch.to(device=self.image.device)

        # Ensure that input batch dtype match the target dtyle of the accumulator
        if batch.dtype != self.image.dtype:
            batch = batch.type_as(self.image)

        for tile, (x, y, tile_width, tile_height) in zip(batch, crop_coords):
            self.image[:, y : y + tile_height, x : x + tile_width] += (
                tile * self.weight
            )
            self.norm_mask[
                :, y : y + tile_height, x : x + tile_width
            ] += self.weight

    @property
    def device(self) -> torch.device:
        return self.image.device

    def merge(self) -> torch.Tensor:
        return self.image / self.norm_mask

    def merge_(self) -> torch.Tensor:
        self.image /= self.norm_mask
        return self.image



#!/usr/bin/env python3
"""
Created on Wed Jan 26 14:33:16 2022
taken from stardist
@author: varunkapoor
"""

import numpy as np

from numba import jit
from tqdm import tqdm
from scipy.optimize import linear_sum_assignment
from collections import namedtuple
from csbdeep.utils import _raise

matching_criteria = dict()


def label_are_sequential(y):
    """returns true if y has only sequential labels from 1..."""
    labels = np.unique(y)
    return (set(labels) - {0}) == set(range(1, 1 + labels.max()))


def is_array_of_integers(y):
    return isinstance(y, np.ndarray) and np.issubdtype(y.dtype, np.integer)


def _check_label_array(y, name=None, check_sequential=False):
    err = ValueError(
        "{label} must be an array of {integers}.".format(
            label="labels" if name is None else name,
            integers=("sequential " if check_sequential else "")
            + "non-negative integers",
        )
    )
    is_array_of_integers(y) or _raise(err)
    if len(y) == 0:
        return True
    if check_sequential:
        label_are_sequential(y) or _raise(err)
    else:
        y.min() >= 0 or _raise(err)
    return True


def label_overlap(x, y, check=True):
    if check:
        _check_label_array(x, "x", True)
        _check_label_array(y, "y", True)
        x.shape == y.shape or _raise(ValueError("x and y must have the same shape"))
    return _label_overlap(x, y)


@jit(nopython=True)
def _label_overlap(x, y):
    x = x.ravel()
    y = y.ravel()
    overlap = np.zeros((1 + x.max(), 1 + y.max()), dtype=np.uint)
    for i in range(len(x)):
        overlap[x[i], y[i]] += 1
    return overlap


def _safe_divide(x, y, eps=1e-10):
    """computes a safe divide which returns 0 if y is zero"""
    if np.isscalar(x) and np.isscalar(y):
        return x / y if np.abs(y) > eps else 0.0
    else:
        out = np.zeros(np.broadcast(x, y).shape, np.float32)
        np.divide(x, y, out=out, where=np.abs(y) > eps)
        return out


def intersection_over_union(overlap):
    _check_label_array(overlap, "overlap")
    if np.sum(overlap) == 0:
        return overlap
    n_pixels_pred = np.sum(overlap, axis=0, keepdims=True)
    n_pixels_true = np.sum(overlap, axis=1, keepdims=True)
    return _safe_divide(overlap, (n_pixels_pred + n_pixels_true - overlap))


matching_criteria["iou"] = intersection_over_union


def intersection_over_true(overlap):
    _check_label_array(overlap, "overlap")
    if np.sum(overlap) == 0:
        return overlap
    n_pixels_true = np.sum(overlap, axis=1, keepdims=True)
    return _safe_divide(overlap, n_pixels_true)


matching_criteria["iot"] = intersection_over_true


def intersection_over_pred(overlap):
    _check_label_array(overlap, "overlap")
    if np.sum(overlap) == 0:
        return overlap
    n_pixels_pred = np.sum(overlap, axis=0, keepdims=True)
    return _safe_divide(overlap, n_pixels_pred)


matching_criteria["iop"] = intersection_over_pred


def precision(tp, fp, fn):
    return tp / (tp + fp) if tp > 0 else 0


def recall(tp, fp, fn):
    return tp / (tp + fn) if tp > 0 else 0


def accuracy(tp, fp, fn):
    # also known as "average precision" (?)
    # -> https://www.kaggle.com/c/data-science-bowl-2018#evaluation
    return tp / (tp + fp + fn) if tp > 0 else 0


def f1(tp, fp, fn):
    # also known as "dice coefficient"
    return (2 * tp) / (2 * tp + fp + fn) if tp > 0 else 0


def matching(y_true, y_pred, thresh=0.5, criterion="iou", report_matches=False):
    """Calculate detection/instance segmentation metrics between ground truth and predicted label images.
    Currently, the following metrics are implemented:
    'fp', 'tp', 'fn', 'precision', 'recall', 'accuracy', 'f1', 'criterion', 'thresh', 'n_true', 'n_pred', 'mean_true_score', 'mean_matched_score', 'panoptic_quality'
    Corresponding objects of y_true and y_pred are counted as true positives (tp), false positives (fp), and false negatives (fn)
    whether their intersection over union (IoU) >= thresh (for criterion='iou', which can be changed)
    * mean_matched_score is the mean IoUs of matched true positives
    * mean_true_score is the mean IoUs of matched true positives but normalized by the total number of GT objects
    * panoptic_quality defined as in Eq. 1 of Kirillov et al. "Panoptic Segmentation", CVPR 2019
    Parameters
    ----------
    y_true: ndarray
        ground truth label image (integer valued)
    y_pred: ndarray
        predicted label image (integer valued)
    thresh: float
        threshold for matching criterion (default 0.5)
    criterion: string
        matching criterion (default IoU)
    report_matches: bool
        if True, additionally calculate matched_pairs and matched_scores (note, that this returns even gt-pred pairs whose scores are below  'thresh')
    Returns
    -------
    Matching object with different metrics as attributes
    Examples
    --------
    >>> y_true = np.zeros((100,100), np.uint16)
    >>> y_true[10:20,10:20] = 1
    >>> y_pred = np.roll(y_true,5,axis = 0)
    >>> stats = matching(y_true, y_pred)
    >>> print(stats)
    Matching(criterion='iou', thresh=0.5, fp=1, tp=0, fn=1, precision=0, recall=0, accuracy=0, f1=0, n_true=1, n_pred=1, mean_true_score=0.0, mean_matched_score=0.0, panoptic_quality=0.0)
    """
    _check_label_array(y_true, "y_true")
    _check_label_array(y_pred, "y_pred")
    y_true.shape == y_pred.shape or _raise(
        ValueError(
            f"y_true ({y_true.shape}) and y_pred ({y_pred.shape}) have different shapes"
        )
    )
    criterion in matching_criteria or _raise(
        ValueError("Matching criterion '%s' not supported." % criterion)
    )
    if thresh is None:
        thresh = 0
    thresh = float(thresh) if np.isscalar(thresh) else map(float, thresh)

    y_true, _, map_rev_true = relabel_sequential(y_true)
    y_pred, _, map_rev_pred = relabel_sequential(y_pred)

    overlap = label_overlap(y_true, y_pred, check=False)
    scores = matching_criteria[criterion](overlap)
    assert 0 <= np.min(scores) <= np.max(scores) <= 1

    # ignoring background
    scores = scores[1:, 1:]
    n_true, n_pred = scores.shape
    n_matched = min(n_true, n_pred)

    def _single(thr):
        not_trivial = n_matched > 0 and np.any(scores >= thr)
        if not_trivial:
            # compute optimal matching with scores as tie-breaker
            costs = -(scores >= thr).astype(float) - scores / (2 * n_matched)
            true_ind, pred_ind = linear_sum_assignment(costs)
            assert n_matched == len(true_ind) == len(pred_ind)
            match_ok = scores[true_ind, pred_ind] >= thr
            tp = np.count_nonzero(match_ok)
        else:
            tp = 0
        fp = n_pred - tp
        fn = n_true - tp
        # assert tp+fp == n_pred
        # assert tp+fn == n_true

        # the score sum over all matched objects (tp)
        sum_matched_score = (
            np.sum(scores[true_ind, pred_ind][match_ok]) if not_trivial else 0.0
        )

        # the score average over all matched objects (tp)
        mean_matched_score = _safe_divide(sum_matched_score, tp)
        # the score average over all gt/true objects
        mean_true_score = _safe_divide(sum_matched_score, n_true)
        panoptic_quality = _safe_divide(sum_matched_score, tp + fp / 2 + fn / 2)

        stats_dict = dict(
            criterion=criterion,
            thresh=thr,
            fp=fp,
            tp=tp,
            fn=fn,
            precision=precision(tp, fp, fn),
            recall=recall(tp, fp, fn),
            accuracy=accuracy(tp, fp, fn),
            f1=f1(tp, fp, fn),
            n_true=n_true,
            n_pred=n_pred,
            mean_true_score=mean_true_score,
            mean_matched_score=mean_matched_score,
            panoptic_quality=panoptic_quality,
        )
        if bool(report_matches):
            if not_trivial:
                stats_dict.update(
                    # int() to be json serializable
                    matched_pairs=tuple(
                        (int(map_rev_true[i]), int(map_rev_pred[j]))
                        for i, j in zip(1 + true_ind, 1 + pred_ind)
                    ),
                    matched_scores=tuple(scores[true_ind, pred_ind]),
                    matched_tps=tuple(map(int, np.flatnonzero(match_ok))),
                )
            else:
                stats_dict.update(
                    matched_pairs=(),
                    matched_scores=(),
                    matched_tps=(),
                )
        return namedtuple("Matching", stats_dict.keys())(*stats_dict.values())

    return _single(thresh) if np.isscalar(thresh) else tuple(map(_single, thresh))


def matching_dataset(
    y_true,
    y_pred,
    thresh=0.5,
    criterion="iou",
    by_image=False,
    show_progress=True,
    parallel=False,
):
    """matching metrics for list of images, see `stardist.matching.matching`"""
    len(y_true) == len(y_pred) or _raise(
        ValueError("y_true and y_pred must have the same length.")
    )
    return matching_dataset_lazy(
        tuple(zip(y_true, y_pred)),
        thresh=thresh,
        criterion=criterion,
        by_image=by_image,
        show_progress=show_progress,
        parallel=parallel,
    )


def matching_dataset_lazy(
    y_gen,
    thresh=0.5,
    criterion="iou",
    by_image=False,
    show_progress=True,
    parallel=False,
):

    expected_keys = {
        "fp",
        "tp",
        "fn",
        "precision",
        "recall",
        "accuracy",
        "f1",
        "criterion",
        "thresh",
        "n_true",
        "n_pred",
        "mean_true_score",
        "mean_matched_score",
        "panoptic_quality",
    }

    single_thresh = False
    if np.isscalar(thresh):
        single_thresh = True
        thresh = (thresh,)

    tqdm_kwargs = {}
    tqdm_kwargs["disable"] = not bool(show_progress)
    if int(show_progress) > 1:
        tqdm_kwargs["total"] = int(show_progress)

    # compute matching stats for every pair of label images
    stats_all = tuple(
        matching(y_t, y_p, thresh=thresh, criterion=criterion, report_matches=False)
        for y_t, y_p in tqdm(y_gen, **tqdm_kwargs)
    )

    # accumulate results over all images for each threshold separately
    n_images, n_threshs = len(stats_all), len(thresh)
    accumulate = [{} for _ in range(n_threshs)]
    for stats in stats_all:
        for i, s in enumerate(stats):
            acc = accumulate[i]
            for k, v in s._asdict().items():
                if k == "mean_true_score" and not bool(by_image):
                    # convert mean_true_score to "sum_matched_score"
                    acc[k] = acc.setdefault(k, 0) + v * s.n_true
                else:
                    try:
                        acc[k] = acc.setdefault(k, 0) + v
                    except TypeError:
                        pass

    # normalize/compute 'precision', 'recall', 'accuracy', 'f1'
    for thr, acc in zip(thresh, accumulate):
        set(acc.keys()) == expected_keys or _raise(ValueError("unexpected keys"))
        acc["criterion"] = criterion
        acc["thresh"] = thr
        acc["by_image"] = bool(by_image)
        if bool(by_image):
            for k in (
                "precision",
                "recall",
                "accuracy",
                "f1",
                "mean_true_score",
                "mean_matched_score",
                "panoptic_quality",
            ):
                acc[k] /= n_images
        else:
            tp, fp, fn, n_true = acc["tp"], acc["fp"], acc["fn"], acc["n_true"]
            sum_matched_score = acc["mean_true_score"]

            mean_matched_score = _safe_divide(sum_matched_score, tp)
            mean_true_score = _safe_divide(sum_matched_score, n_true)
            panoptic_quality = _safe_divide(sum_matched_score, tp + fp / 2 + fn / 2)

            acc.update(
                precision=precision(tp, fp, fn),
                recall=recall(tp, fp, fn),
                accuracy=accuracy(tp, fp, fn),
                f1=f1(tp, fp, fn),
                mean_true_score=mean_true_score,
                mean_matched_score=mean_matched_score,
                panoptic_quality=panoptic_quality,
            )

    accumulate = tuple(
        namedtuple("DatasetMatching", acc.keys())(*acc.values()) for acc in accumulate
    )
    return accumulate[0] if single_thresh else accumulate


# copied from scikit-image master for now (remove when part of a release)
def relabel_sequential(label_field, offset=1):
    """Relabel arbitrary labels to {`offset`, ... `offset` + number_of_labels}.
    This function also returns the forward map (mapping the original labels to
    the reduced labels) and the inverse map (mapping the reduced labels back
    to the original ones).
    Parameters
    ----------
    label_field : numpy array of int, arbitrary shape
        An array of labels, which must be non-negative integers.
    offset : int, optional
        The return labels will start at `offset`, which should be
        strictly positive.
    Returns
    -------
    relabeled : numpy array of int, same shape as `label_field`
        The input label field with labels mapped to
        {offset, ..., number_of_labels + offset - 1}.
        The data type will be the same as `label_field`, except when
        offset + number_of_labels causes overflow of the current data type.
    forward_map : numpy array of int, shape ``(label_field.max() + 1,)``
        The map from the original label space to the returned label
        space. Can be used to re-apply the same mapping. See examples
        for usage. The data type will be the same as `relabeled`.
    inverse_map : 1D numpy array of int, of length offset + number of labels
        The map from the new label space to the original space. This
        can be used to reconstruct the original label field from the
        relabeled one. The data type will be the same as `relabeled`.
    Notes
    -----
    The label 0 is assumed to denote the background and is never remapped.
    The forward map can be extremely big for some inputs, since its
    length is given by the maximum of the label field. However, in most
    situations, ``label_field.max()`` is much smaller than
    ``label_field.size``, and in these cases the forward map is
    guaranteed to be smaller than either the input or output images.
    Examples
    --------
    >>> from skimage.segmentation import relabel_sequential
    >>> label_field = np.array([1, 1, 5, 5, 8, 99, 42])
    >>> relab, fw, inv = relabel_sequential(label_field)
    >>> relab
    array([1, 1, 2, 2, 3, 5, 4])
    >>> fw
    array([0, 1, 0, 0, 0, 2, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0,
           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5])
    >>> inv
    array([ 0,  1,  5,  8, 42, 99])
    >>> (fw[label_field] == relab).all()
    True
    >>> (inv[relab] == label_field).all()
    True
    >>> relab, fw, inv = relabel_sequential(label_field, offset=5)
    >>> relab
    array([5, 5, 6, 6, 7, 9, 8])
    """
    offset = int(offset)
    if offset <= 0:
        raise ValueError("Offset must be strictly positive.")
    if np.min(label_field) < 0:
        raise ValueError("Cannot relabel array that contains negative values.")
    max_label = int(label_field.max())  # Ensure max_label is an integer
    if not np.issubdtype(label_field.dtype, np.integer):
        new_type = np.min_scalar_type(max_label)
        label_field = label_field.astype(new_type)
    labels = np.unique(label_field)
    labels0 = labels[labels != 0]
    new_max_label = offset - 1 + len(labels0)
    new_labels0 = np.arange(offset, new_max_label + 1)
    output_type = label_field.dtype
    required_type = np.min_scalar_type(new_max_label)
    if np.dtype(required_type).itemsize > np.dtype(label_field.dtype).itemsize:
        output_type = required_type
    forward_map = np.zeros(max_label + 1, dtype=output_type)
    forward_map[labels0] = new_labels0
    inverse_map = np.zeros(new_max_label + 1, dtype=output_type)
    inverse_map[offset:] = labels0
    relabeled = forward_map[label_field]
    return relabeled, forward_map, inverse_map
